/*
 * ar3cmonitor - cdata.cpp
 *
 * Copyright (C) 2006..2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <qpainter.h>
#include <cmath>
#include <niceformat.h>

#include "ameter.h"
#include "control.h"
#include "ar3cm_proto.h"
#include <cdata.h>

MeterAWidget::MeterAWidget(int _analogIndex, QColor _color, float _maxValue,
		float _minValue, float _unitValue, int _angle, int _scale,
		QString _caption, QString _unit,
		unsigned long flags, prec_t _prec, QWidget *parent,
		const char *name, Qt::WFlags f)
	: BufferedWidget(parent, name, f), caption(_caption), unit(_unit),
	prec(_prec)
{
	setMinimumHeight(60);
	setMinimumWidth(60);
	max_value = _maxValue;
	min_value = _minValue;
	unit_value = _unitValue;
	color = _color;
	scale = _scale;
	angle = M_PI * _angle / (180.0 * 16);
	unipolar = flags & AR3CM_AMETER_UNIPOLAR;
	show_value = flags & AR3CM_AMETER_VALUE;
	along = flags & AR3CM_AMETER_ALONG;

	analogIndex = _analogIndex;
	oldvalid = 0;
	old_line.x1 = 0;
	old_line.x2 = 0;
	old_line.y1 = 0;
	old_line.y2 = 0;
	old_niceFormat = "";
}

QString MeterAWidget::niceFormat(float v)
{
	char buf[32];

	niceformat(buf, sizeof(buf), unit_value, min_value, max_value, v,
			prec);
	return QString(buf);
}

void MeterAWidget::paint(QPainter *p)
{
	int h = height(), w = width();
	int cx, cy;
	int i;
	int text_h, text_w;
	int iv;
	int ret;

	QPen pen(p->pen());

	if (along)
		ret = cdata_get_along(analogIndex, &iv);
	else
		ret = cdata_get_analog(analogIndex, &iv);

	if (ret) {
		p->setPen(Qt::red);
		p->drawText(0, 0, width(), height(), Qt::AlignCenter,
				QString("?"));
		oldvalid = 0;
		return;
	}
	float v = iv / unit_value;

	if (unipolar)
		v = fabs(v);

	oldvalid = 1;
	old_value = v;

	float range = ceilf(max(fabs(min_value), fabs(max_value)) / scale)
		* scale;

	float a_offset = unipolar ? -angle / 2.0 : 0;
	int px, py;

	float a = angle * (unipolar ? 0.5 : 1.0);
	float ratio = max(1, 1 - cos(a));

	cx = width() / 2;
	cy = (int) round((height() - 20) / ratio);

	h -= 20;
	w -= 20;

	if (unipolar)
		h = (int) round(h * (ratio + 1));

	if (h > w)
		h = w;
	else
		w = h;

	pen.setColor(QColor(0, 0, 170, QColor::Hsv));
	pen.setWidth(2);
	p->setPen(pen);

	QFont f(CO->valueFont());
	f.setPointSize(8);
	p->setFont(f);
	text_h = QFontMetrics(f).height();
	text_w = QFontMetrics(f).width(QString("9999"));

	p->drawText(0, cy + text_h / 2, width(), text_h,
			Qt::AlignHCenter | Qt::AlignCenter | Qt::SingleLine,
			QString("%1 [%2]").arg(caption).arg(unit));
	if (show_value) {
		p->drawText(0, cy + 3 * text_h / 2, width(), text_h,
				Qt::AlignHCenter | Qt::AlignCenter
				| Qt::SingleLine,
				QString("%1").arg(niceFormat(v)));
	}

	for (i = unipolar ? 0 : -scale; i <= scale; i++) {
		float a = a_offset + ((float) i / (float) scale) * angle;
		float xs = sin(a);
		float ys = -cos(a);
		pen.setColor(QColor(0, 0, 170, QColor::Hsv));
		p->setPen(pen);

		p->drawLine(cx + (int) round(xs * (w / 3.0)),
				cy + (int) round(ys * (h / 3.0)),
				cx + (int) round(xs * (w / 2.6)),
				cy + (int) round(ys * (h / 2.6)));

		p->drawText(cx + (int) round(xs * (w / 2.2)) - text_w/2,
				cy + (int) round(ys * (h / 2.2)) - text_h/2,
				text_w, text_h,
			Qt::AlignHCenter | Qt::AlignCenter | Qt::SingleLine,
			QString().sprintf("%d",
				(int) round((i / (float) scale) * range)));
	}

	px = cx + (int) round((w / 2.3) * sin(a_offset + angle * v / range));
	py = cy - (int) round((h / 2.3) * cos(a_offset + angle * v / range));

	old_line.x1 = cx;
	old_line.y1 = cy;
	old_line.x2 = px;
	old_line.y2 = py;
	old_niceFormat = QString("%1").arg(niceFormat(v));

	pen.setColor(color);
	pen.setWidth(3);
	p->setPen(pen);
	p->setBrush(color);
	p->drawLine(cx, cy, px, py);
}

struct line MeterAWidget::resolveLinePos(float v)
{
	struct line ln;
	int h = height() - 20;
	int w = width() - 20;

	float a = angle * (unipolar ? 0.5 : 1.0);
	float ratio = max(1, 1 - cos(a));
	float a_offset = unipolar ? -angle / 2.0 : 0;
	float range = ceilf(max(fabs(min_value), fabs(max_value)) / scale)
		* scale;

	ln.x1 = width() / 2;
	ln.y1 = (int) round((height() - 20) / ratio);

	ln.x2 = ln.x1 + (int) round((w / 2.3) * sin(a_offset + angle * v
				/ range));
	ln.y2 = ln.y1 - (int) round((h / 2.3) * cos(a_offset + angle * v
				/ range));

	return ln;
}

void MeterAWidget::updateMeter()
{
	int need_update = 0;
	int iv;

	if (cdata_get_analog(analogIndex, &iv)) {
		if (oldvalid)
			update();
		return;
	}

	if (oldvalid) {
		float v = iv / unit_value;
		struct line cur = resolveLinePos(v);

		if (old_line.x2 != cur.x2) {
			need_update = 1;
		} else if (old_line.y2 != cur.y2) {
			need_update = 1;
		} else if (old_niceFormat != QString("%1").arg(niceFormat(v))) {
			need_update = 1;
		}
	} else {
		need_update = 1;
	}

	if (need_update)
		update();
}
