/*
 * ar3cmonitor - offstore.h
 *
 * 	Offline Data Store
 * 
 * Copyright (C) 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef OFFSTORE_H_INCLUDED
#define OFFSTORE_H_INCLUDED

#include "datastore.h"

class OfflineBuffer : public DataStore {
public:
	OfflineBuffer *tnext;
	OfflineBuffer *prev, *next;
	unsigned long minute;
	int completed;
	
	OfflineBuffer(int binaries, unsigned long min, const char *name = NULL);
	virtual int shrink(int prio);
	void garbageCollector();
	virtual ~OfflineBuffer();
};

class OfflineBufferListSentinel : public OfflineBuffer {
public:
	OfflineBufferListSentinel()
		: OfflineBuffer(0, 0, "Sentinel")
	{
		prev = this;
		next = this;
	}
	
	virtual int shrink()
	{
		return 0;
	}
	
	~OfflineBufferListSentinel()
	{
	}
};

class OfflineBufferList {
private:
	OfflineBufferListSentinel Sentinel;
	unsigned int count;
	
public:
	OfflineBufferList();
	
	unsigned int get_count()
	{
		return count;
	}

	OfflineBuffer *get_head()
	{
		return count ? Sentinel.next : NULL;
	}
	
	unsigned long get_used_all();

	void push(OfflineBuffer *b)
	{
		b->prev = Sentinel.prev;
		b->next = &Sentinel;
		Sentinel.prev->next = b;
		Sentinel.prev = b;
		count++;
	}

	OfflineBuffer *pop()
	{
		OfflineBuffer *b;
		
		if (!count)
			return NULL;
		b = Sentinel.next;
		Sentinel.next = b->next;
		Sentinel.next->prev = &Sentinel;
		count--;
		return b;
	}

	void unlink(OfflineBuffer *b)
	{
		b->prev->next = b->next;
		b->next->prev = b->prev;
		count--;
	}

	bool empty()
	{
		return Sentinel.next == &Sentinel;
	}
	
//	void clear();

	~OfflineBufferList();
};

class OfflineBufferQueue {
private:
	OfflineBuffer *head, *tail;
	
public:
	OfflineBufferQueue()
		: head(NULL)
	{
	}

	void push(OfflineBuffer *b)
	{
		b->tnext = NULL;
		if (head != NULL)
			tail->tnext = b;
		else
			head = b;
		tail = b;
	}
	
	void push_head(OfflineBuffer *b)
	{
		b->tnext = head;
		if (head == NULL)
			tail = b;
		head = b;
	}

	OfflineBuffer *get_head()
	{
		return head;
	}
	
	OfflineBuffer *pop()
	{
		OfflineBuffer *ret;

		ret = head;
		if (head != NULL)
			head = head->tnext;
		return ret;
	}

	void clear()
	{
		while (head != NULL)
			pop();
	}

	~OfflineBufferQueue()
	{
		clear();
	}
};

/*
 * Obiekt ten troszczy si o pobieranie danych w trybie offline.
 */
class OfflineStore : public DataStore {
private:
	unsigned long LMin, RMin;	/* current download pointers */
	unsigned long MLMin, MRMin;	/* merged pointers */
	OfflineBufferList requests;
	OfflineBufferQueue left, right;
	unsigned int req, req_limit;
	unsigned int t_counter;
	int wait_wdt;	/* wait for watchdog, used to resync */
	int prefetch;
	int do_resync;

	bool buffer_available()
	{
		return requests.get_count() < req_limit;
	}
	
	void buffer_delete(OfflineBuffer *b)
	{
		delete b;
	}
	
	OfflineBuffer *buffer_get_current()
	{
		return requests.get_head();
	}
	
	void buffer_finalize(OfflineBuffer *b);
	unsigned int calc_prefetch(unsigned int base, unsigned int ratio,
			unsigned long long range);
	void download_left();
	void download_right();
	int __download();
	void try_merge_left();
	void try_merge_right();
	void request_lmin_high();
	void request_rmin_high();
	void disable_prefetch()
	{
		prefetch = 0;
	}

	void enable_prefetch()
	{
		prefetch = 1;
	}
	void remove_left();
	void remove_right();
	void request_abort(void);

public:
	OfflineStore(unsigned int nBinaries, const char *_name = NULL);

	void set_req_limit(unsigned int limit)
	{
		req_limit = limit;
	}

	unsigned long get_used_all();
	unsigned long get_used_bufs();
	
	int shrink(int prio);
	
	int shrink_buffers(int prio);	

	void try_merge();

	void download();
	void newBRecord(unsigned long long ts, int id, BState state);
	void newARecord(unsigned long long ts, ARecord *r);
	void beginUndefined(unsigned long long ts, unsigned int flags);
	
	void watchdog();
	void please_update();
	void clear_reqs();

	void finish_minute(); 
	void garbageCollector();

	void clear();
	void connected();
	void resync();

	~OfflineStore();
};

#endif
