/*
 * ar3cmonitor - timestampedlist.h
 * 
 * Copyright (C) 2004  Szymon Acedanski <accek@podlesie.net>
 * Copyright (C) 2005..2011  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Fundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef TIMESTAMPEDLIST_H
#define TIMESTAMPEDLIST_H

#include <stdio.h>
#include <stdlib.h>

#include "consts.h"
#include "ar3cmonitor.h"
#include "monalloc.h"

class Item;

#ifdef TSLIST_STATS
extern unsigned long sentinels, items, iters;

#define iters_inc()	do { iters++; } while (0)
#define iters_dec()	do { iters--; } while (0)

#define items_inc()	do { items++; } while (0)
#define items_dec()	do { items--; } while (0)

#define sentinels_inc()	do { sentinels++; } while (0)
#define sentinels_dec()	do { sentinels--; } while (0)

#else

#define iters_inc()	do { } while (0)
#define iters_dec()	do { } while (0)

#define items_inc()	do { } while (0)
#define items_dec()	do { } while (0)

#define sentinels_inc()	do { } while (0)
#define sentinels_dec()	do { } while (0)
#endif

class UIterator {
protected:
	Item *Ptr;

public:
	UIterator()
		: Ptr(NULL)
	{
		iters_inc();
	}

	UIterator(const UIterator &i)
		: Ptr(i.Ptr)
	{
		iters_inc();
	}

	explicit UIterator(Item *p)
		: Ptr(p)
	{
		iters_inc();
	}

	virtual ~UIterator()
	{
		iters_dec();
	}

	Item *get_item()
	{
		return Ptr;
	}

	bool operator ==(const UIterator & b)
	{
		return Ptr == b.Ptr;
	}

	bool operator ==(const Item *b)
	{
		return Ptr == b;
	}

};

class Item : public monalloc {
public:
	Item *Next;
	Item *Prev;
	unsigned long long TimeStamp;
	int valid;

	Item(Item *Next = NULL, Item *Prev = NULL,
			unsigned long long TimeStamp = 0, int is_valid = 1)
		: Next(Next), Prev(Prev), TimeStamp(TimeStamp), valid(is_valid)
	{
		items_inc();
	}

	virtual size_t get_size()
	{
		abort();
	}

	virtual ~Item()
	{
		items_dec();
	}
};

template < class T > class TimestampedList {
public:
	class Iterator : public UIterator {
	public:
		Iterator()
			: UIterator()
		{
		}

		Iterator(const UIterator &i)
			: UIterator(i)
		{
		}

		Iterator(const Iterator &i)
			: UIterator(i)
		{
		}

		explicit Iterator(Item *p)
			: UIterator(p)
		{
		}

		virtual ~Iterator()
		{
		}

		Iterator operator +(int a)
		{
			Iterator ret(*this);
			int i;

			for (i = 0; i < a; i++)
				++ret;
			return ret;
		}

		Iterator operator -(int a)
		{
			Iterator ret(*this);
			int i;

			for (i = 0; i < a; i++)
				--ret;
			return ret;
		}

		Iterator &operator ++()
		{
			Ptr = Ptr->Next;
			return *this;
		}

		Iterator &operator --()
		{
			Ptr = Ptr->Prev;
			return *this;
		}

		void destroyData()
		{
			delete Ptr;

			Ptr = NULL;
		}

		T *data()
		{
			return (T *) Ptr;
		}

		unsigned long long timeStamp()
		{
			return Ptr->TimeStamp;
		}

		size_t get_size()
		{
			return Ptr->get_size();
		}

		bool valid()
		{
			return Ptr->valid;
		}

		void setTimeStamp(unsigned long long ts)
		{
			Ptr->TimeStamp = ts;
		}

		void seekLeftBefore(unsigned long long ts)
		{
			while (valid() && timeStamp() > ts)
				Ptr = Ptr->Prev;
		}

		static Iterator scanLeftBefore(unsigned long long ts,
				Iterator it)
		{
			while (it.valid() && it.timeStamp() > ts)
				--it;
			return it;
		}

		Iterator scanLeftBefore(unsigned long long ts)
		{
			return scanLeftBefore(ts, *this);
		}

		static Iterator scanLeftAfter(unsigned long long ts,
				Iterator it)
		{
			it = scanLeftBefore(ts, it);
			if (it.timeStamp() < ts)
				++it;
			return it;
		}

		Iterator scanLeftAfter(unsigned long long ts)
		{
			return scanLeftAfter(ts, *this);
		}

		static Iterator scanRightAfter(unsigned long long ts,
				Iterator it)
		{
			while (it.valid() && it.timeStamp() < ts)
				++it;
			return it;
		}

		Iterator scanRightAfter(unsigned long long ts)
		{
			return scanRightAfter(ts, *this);
		}

		static Iterator scanRightBefore(unsigned long long ts,
				Iterator it)
		{
			it = scanRightAfter(ts, it);
			if (it.timeStamp() > ts)
				--it;
			return it;
		}

		Iterator scanRightBefore(unsigned long long ts)
		{
			return scanRightBefore(ts, *this);
		}

		Iterator scanAfter(unsigned long long ts)
		{
			Iterator it(*this);

			if (!it.valid())
				--it;

			if (!it.valid())
				return it;

			return (it.timeStamp() > ts) ? scanLeftAfter(ts, it)
				: scanRightAfter(ts, it);
		}

		Iterator scanBefore(unsigned long long ts)
		{
			Iterator it(*this);

			if (!it.valid())
				--it;

			if (!it.valid())
				return it;

			return (it.timeStamp() > ts) ? scanLeftBefore(ts, it)
				: scanRightBefore(ts, it);
		}

		Iterator insertAfter(unsigned long long TimeStamp, Item *i)
		{
			unsigned long long t = timeStamp();

			if (TimeStamp < t) {
				qWarning("warning: non-monotonic timestamp: "
						"%lld -> %lld, fixing it",
						TimeStamp, t);
				TimeStamp = t;
			}
			i->Prev = Ptr;
			i->Next = Ptr->Next;
			i->TimeStamp = TimeStamp;
			i->Next->Prev = i;
			Ptr->Next = i;
			return Iterator(i);
		}

	};

private:
	Item *Sentinel;

	void detach()
	{
		Sentinel->Next = Sentinel;
		Sentinel->Prev = Sentinel;
	}

public:
	TimestampedList()
	{
		Sentinel = new Item(NULL, NULL, 0, 0);
		sentinels_inc();
		Q_ASSERT(Sentinel);
		Sentinel->Next = Sentinel;
		Sentinel->Prev = Sentinel;
	}

	~TimestampedList()
	{
		clear();
		sentinels_dec();
		delete Sentinel;
	}

	void clear()
	{
		Iterator n = ++it();

		while (n.valid()) {
			Iterator i = n;
			++n;
			i.destroyData();
		}
		detach();
	}

	void pushBack(unsigned long long TimeStamp, Item *i)
	{
		i->Prev = Sentinel->Prev;
		i->Next = Sentinel;
		i->TimeStamp = TimeStamp;
		Sentinel->Prev->Next = i;
		Sentinel->Prev = i;
	}

	void pushFront(unsigned long long TimeStamp, Item *i)
	{
		i->Prev = Sentinel;
		i->Next = Sentinel->Next;
		i->TimeStamp = TimeStamp;
		Sentinel->Next->Prev = i;
		Sentinel->Next = i;
	}

	T *head(void)
	{
		return (T *) Sentinel->Next;
	}

	T *tail(void)
	{
		return (T *) Sentinel->Prev;
	}

	T *poph(void)
	{
		Item *First = Sentinel->Next;

		Sentinel->Next = First->Next;
		Sentinel->Next->Prev = Sentinel;
		return (T *) First;
	}

	T *popt(void)
	{
		Item *Last = Sentinel->Prev;

		Sentinel->Prev = Last->Prev;
		Sentinel->Prev->Next = Sentinel;
		return (T *) Last;
	}

	void popFront()
	{
		Item *First = Sentinel->Next;
		Sentinel->Next = First->Next;
		Sentinel->Next->Prev = Sentinel;
		delete First;
	}

	void popBack()
	{
		Item *Last = Sentinel->Prev;
		Sentinel->Prev = Last->Prev;
		Sentinel->Prev->Next = Sentinel;
		delete Last;
	}

	void left_join(TimestampedList < T > &list)
	{
		Item *i, *t;

		if (list.empty())
			return;

		i = list.get_list();
		t = Sentinel->Next;

		t->Prev = i->Prev;
		Sentinel->Next = i->Next;

		Sentinel->Next->Prev = Sentinel;
		t->Prev->Next = t;

		list.put_list(i);
	}

	void right_join(TimestampedList < T > &list)
	{
		Item *i, *t;

		if (list.empty())
			return;

		i = list.get_list();
		t = Sentinel->Prev;

		t->Next = i->Next;
		Sentinel->Prev = i->Prev;

		t->Next->Prev = t;
		Sentinel->Prev->Next = Sentinel;

		list.put_list(i);
	}

	bool empty()
	{
		if ((Sentinel->Next == Sentinel)) {
			if (Sentinel->Prev != Sentinel)
				abort();
		}
		return Sentinel->Next == Sentinel;
	}

	bool tmore()
	{
		return Sentinel->Next != Sentinel->Prev;
	}

	Item *get_list()
	{
		Item *ret = Sentinel;

		Sentinel = new Item(NULL, NULL, 0, 0);
		Q_ASSERT(Sentinel);
		sentinels_inc();
		Sentinel->Next = Sentinel;
		Sentinel->Prev = Sentinel;
		return ret;
	}

	void put_list(Item *s)
	{
		delete s;
		sentinels_dec();
	}

	Iterator it()
	{
		return Iterator(Sentinel);
	}
};

#endif
